from abc import ABC, abstractmethod
from functools import partialmethod
from typing import Any, Type, Dict, Tuple


class BaseProvider(ABC):
    _instance: Any = None
    _args: Tuple[Any, ...]
    _kwargs: Dict[str, Any]

    def __init__(self, provided_class: Type, *args, **kwargs):
        """
        Base class of all providers
        :param provided_class: class that the provider provides
        :param args: constructor args of the provided class
        :param kwargs: constructor kwargs of the provided class
        """
        
        self._args = args
        self._kwargs = kwargs
        self._provided_class = provided_class

    @abstractmethod
    def get_instance(self) -> Any: # pragma: no cover
        """
        Get an instance of the provided class
        :return: instance of the provided class
        """
        pass

    @abstractmethod
    def wire(self): # parma: no cover
        """
        Wire the provided class and its dependencies
        :return:
        """
        pass # pragma: no cover

    @property
    def provided_class_name(self):
        """
        Getter for the provided class name
        :return: name of the provided class
        """
        return self._provided_class.__name__

    @property
    def provided_class(self):
        return self._provided_class

    def inject_provided_init(self, patched_init: partialmethod): # pragma: no cover
        self._provided_class.__init__ = patched_init


class FactoryProvider(BaseProvider):
    """
    Provides a new injectable every time it is injected
    """
    def get_instance(self) -> Any:
        assert self._instance is not None, "Container is not wired yet"

        return self._instance(*self._args, **self._kwargs)

    def wire(self):
        # TODO: resolve the dependencies of the provided class
        self._instance = self._provided_class


class SingletonProvider(BaseProvider):
    """
    Provides a singleton injectable
    """
    def get_instance(self) -> Any:
        return self._instance

    def wire(self):
        self._instance = self._provided_class(*self._args, **self._kwargs)



import functools
import inspect
import typing

from starlette.middleware import Middleware
from starlette.routing import Route, get_name, request_response, compile_path

from adnexus.wrappers import InjectedCallable


# this is necessary because the standard starlette route would handle our injected objects as an Endpoint (which does not work)
class InjectedRoute(Route):
    def __init__(
        self,
        path: str,
        endpoint: typing.Callable[..., typing.Any],
        *,
        methods: list[str] | None = None,
        name: str | None = None,
        include_in_schema: bool = True,
        middleware: typing.Sequence[Middleware] | None = None,
    ) -> None:
        assert path.startswith("/"), "Routed paths must start with '/'"
        self.path = path
        self.endpoint = endpoint
        self.name = get_name(endpoint) if name is None else name
        self.include_in_schema = include_in_schema

        endpoint_handler = endpoint
        while isinstance(endpoint_handler, functools.partial):
            endpoint_handler = endpoint_handler.func
        if inspect.isfunction(endpoint_handler) or inspect.ismethod(endpoint_handler) or isinstance(endpoint_handler, InjectedCallable):
            # Endpoint is function method or InjectedCallable. Treat it as `func(request) -> response`.
            self.app = request_response(endpoint)
            if methods is None:
                methods = ["GET"]
        else:
            # Endpoint is a class. Treat it as ASGI.
            self.app = endpoint

        if middleware is not None:
            for cls, args, kwargs in reversed(middleware):
                self.app = cls(app=self.app, *args, **kwargs)

        if methods is None:
            self.methods = None
        else:
            self.methods = {method.upper() for method in methods}
            if "GET" in self.methods:
                self.methods.add("HEAD")

        self.path_regex, self.path_format, self.param_convertors = compile_path(path)




def merge(dict1: dict, dict2: dict) -> dict:
    new_dict = {}
    for key, value in dict1.items():
        new_dict[key] = value

    for key, value in dict2.items():
        new_dict[key] = value

    return new_dict



from typing import Type, List
from pydantic import BaseModel
from .base_loader import BaseConfigLoader

from .helpers import merge


def load_config(loaders: List[BaseConfigLoader], model: Type[BaseModel]) -> BaseModel:
    """
    Load config from multiple loaders in a model
    :param loaders: List of config loaders to use
    :param model: config model
    :return: Config
    """
    dict_config = {}
    for loader in loaders:
        dict_config = merge(dict_config, loader.get_config())

    return model.parse_obj(dict_config)



import functools
import importlib
import inspect
from abc import ABC
from typing import List, Dict, get_origin, get_args, Any

from pydantic import BaseModel

from adnexus.wrappers import InjectedCallable, InjectedClass
from adnexus.exceptions import ImproperlyConfigured, WiringError
from adnexus.providers import BaseProvider
from adnexus.markers import Provide


class BaseContainer(ABC):
    config: BaseModel
    injectables: List[BaseProvider]

    config: BaseModel
    _provider_mapping: Dict[str, BaseProvider] = {}

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    def __init__(self):
        """
        Class all container types inherit from
        """
        if len(self.get_injectables()) < 1:
            raise ImproperlyConfigured("Container must have at least one injectable")

    def wire(self, modules: List[str]):
        """
        Parse the config and do implementation specific wiring stuff
        :param modules: Modules of injectable callables
        """
        pass

    def get_injectables(self):
        if hasattr(self, "injectables"):
            return self.injectables
        else:
            return []


class DeclarativeContainer(BaseContainer):
    @staticmethod
    def _get_root_callables(mod_paths: List[str]):
        root_callables = []
        for mod_path in mod_paths:
            module = importlib.import_module(mod_path)
            root_callables.extend(
                [dep[1] for dep in inspect.getmembers(module, predicate=lambda m: isinstance(m, InjectedCallable) or isinstance(m, InjectedClass))])

        return root_callables

    def _resolve(self, injectable_obj: Any):
        if isinstance(injectable_obj, BaseProvider):
            # resolving dependencies of a dependency
            params = inspect.signature(injectable_obj.provided_class.__init__).parameters.items()
        elif isinstance(injectable_obj, InjectedCallable):
            # resolving dependencies of a InjectedCallable or AsyncInjectedCallable
            params = inspect.signature(injectable_obj.wrapped).parameters.items()
        elif isinstance(injectable_obj, InjectedClass):
            # resolving dependencies of InjectedClass
            params = inspect.signature(injectable_obj.wrapped.__init__).parameters.items()

        dependencies = {}
        for param_name, param_type in params:
            if get_origin(param_type.annotation) == Provide:
                dependency_class_name = get_args(param_type.annotation)[0].__name__
                if dependency_class_name not in self._provider_mapping.keys():
                    raise WiringError(f'"{dependency_class_name}" is not registered in container "{self.__class__.__name__}"')

                # resolve dependencies of dependencies
                current_dep_params = {key: value.get_instance() for key, value in self._resolve(self._provider_mapping[dependency_class_name]).items()}
                partial_init = functools.partialmethod(self._provider_mapping[dependency_class_name].provided_class.__init__, **current_dep_params)
                self._provider_mapping[dependency_class_name].inject_provided_init(partial_init)

                dependencies[param_name] = self._provider_mapping[dependency_class_name]

        return dependencies

    def wire(self, modules: List[str]):
        """
        Wire all injectables of this container
        :param modules:
        :return: Modules of injectable callables
        """

        super().wire(modules)

        for provider in self.get_injectables():
            provider.wire()
            self._provider_mapping[provider.provided_class_name] = provider

        for root_callable in self._get_root_callables(modules):
            setattr(root_callable, "dependencies", self._resolve(root_callable))



import inspect
from typing import Callable, Type, Any

from adnexus.wrappers import InjectedCallable, InjectedAsyncCallable, InjectedClass


def inject(wrapped: Callable | Type[Any]):
    """
    Decorator for marking a callable as injectable
    :param wrapped: Callable or Class to decorate
    :return: InjectedCallable[fn]
    """
    if inspect.iscoroutinefunction(wrapped):
        return InjectedAsyncCallable(wrapped)
    elif inspect.isfunction(wrapped):
        return InjectedCallable(wrapped)
    elif inspect.isclass(wrapped):
        return InjectedClass(wrapped)
    else:
        raise TypeError(f"{wrapped} cannot be used as a root injection object")



from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict


class BaseConfigLoader(ABC):
    def __init__(self, *args, **kwargs): # pragma: no cover
        pass

    @abstractmethod
    def get_config(self) -> Dict[str, Any]: # pragma: no cover
        pass


class FileConfigLoader(BaseConfigLoader):
    def __init__(self, path: Path):
        with path.open("r") as f:
            self.data = self.parse(f.read())

    @abstractmethod
    def parse(self, raw_data: str) -> dict:
        pass